"""
Definition of the base feature of any ruledb.
"""
import abc
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from comb_spec_searcher.class_db import ClassDB
from comb_spec_searcher.exception import SpecificationNotFound
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.strategies.strategy_pack import StrategyPack

if TYPE_CHECKING:
    from comb_spec_searcher import CombinatorialSpecificationSearcher

RuleDBMethod = TypeVar("RuleDBMethod", bound=Callable[..., Any])


__all__ = ["ensure_specification", "RuleDBAbstract"]


class RuleDBAbstract(abc.ABC):
    NOT_LINK_ERROR = RuntimeError(
        "RuleDB is not linked with the searcher. Call `link_searcher` before using it."
    )

    def __init__(self) -> None:
        self._searcher: Optional["CombinatorialSpecificationSearcher"] = None

    def link_searcher(self, searcher: "CombinatorialSpecificationSearcher") -> None:
        if self._searcher is not None:
            raise RuntimeError("Searcher is alreay linked")
        self._searcher = searcher

    @property
    def searcher(self) -> "CombinatorialSpecificationSearcher":
        if self._searcher is None:
            raise RuleDBAbstract.NOT_LINK_ERROR
        return self._searcher

    @property
    def classdb(self) -> ClassDB:
        return self.searcher.classdb

    @property
    def strategy_pack(self) -> StrategyPack:
        return self.searcher.strategy_pack

    @property
    def root_label(self) -> int:
        return self.searcher.start_label

    @abc.abstractmethod
    def is_verified(self, label: int) -> bool:
        """Return True if label has been verified."""

    @abc.abstractmethod
    def add(self, start: int, ends: Tuple[int, ...], rule: AbstractRule) -> None:
        """
        Add a rule to the database.

        - start is a single integer.
        - ends is a tuple of integers, representing the children.
        - rule is a Rule that creates start -> ends.
        """

    @abc.abstractmethod
    def status(self, elaborate: bool) -> str:
        """Return a string describing the status of the rule database."""

    @abc.abstractmethod
    def has_specification(self) -> bool:
        """Indicate if the rule db contains a specification for the given label."""

    @abc.abstractmethod
    def get_specification_rules(self, **kwargs) -> Iterator[AbstractRule]:
        """
        Return the rules that constitute a specification.

        Raise a SpecificationNotFound error if the ruledb does not contain a
        specification.

        The kwargs should to used to pass some option for the specification search that
        are specific to each type of ruledb.
        """


def ensure_specification(func: RuleDBMethod) -> RuleDBMethod:
    def inner(ruledb: RuleDBAbstract, *args, **kwargs):
        assert isinstance(ruledb, RuleDBAbstract)
        if not ruledb.has_specification():
            raise SpecificationNotFound
        return func(ruledb, *args, **kwargs)

    return cast(RuleDBMethod, inner)



# pylint: disable=too-many-lines
"""
The rule class is used for a specific application of a strategy on a combinatorial
class. This is not something the user should implement, as it is just a wrapper for
calling the Strategy class and storing its results.
"""


import abc
import random
from collections import defaultdict
from importlib import import_module
from itertools import chain, product
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from sympy import Eq, Function

from comb_spec_searcher.combinatorial_class import CombinatorialClass
from comb_spec_searcher.typing import (
    ForestRuleKey,
    Objects,
    ObjectsCache,
    RuleBucket,
    SubObjects,
    SubTerms,
    Terms,
)

from ..combinatorial_class import CombinatorialClassType, CombinatorialObjectType
from ..exception import SanityCheckFailure, StrategyDoesNotApply
from ..utils import TermsCache
from .constructor import Complement, Constructor, DisjointUnion

if TYPE_CHECKING:
    from .strategy import AbstractStrategy, Strategy, VerificationStrategy
    from .strategy_pack import StrategyPack

__all__ = ("Rule", "VerificationRule")


class AbstractRule(abc.ABC, Generic[CombinatorialClassType, CombinatorialObjectType]):
    """
    An application of a strategy to a comb_class. If the children are not provided they
    are computed from the strategy.
    """

    def __init__(
        self,
        strategy: "AbstractStrategy[CombinatorialClassType, CombinatorialObjectType]",
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ):
        self.comb_class = comb_class
        self._strategy = strategy
        self.terms_cache: TermsCache = TermsCache()
        self.objects_cache: ObjectsCache = []
        self.subrecs: Optional[Tuple[Callable[..., int], ...]] = None
        self.subgenerators: Optional[
            Tuple[Callable[..., Iterator[CombinatorialObjectType]], ...]
        ] = None
        self.subsamplers: Optional[
            Tuple[Callable[..., CombinatorialObjectType], ...]
        ] = None
        self.subterms: Optional[SubTerms] = None
        self.subobjects: Optional[SubObjects] = None
        self._children = children
        self._non_empty_children: Optional[Tuple[CombinatorialClassType, ...]] = None
        self._shifts: Optional[Tuple[int, ...]] = None

    def to_jsonable(self) -> dict:
        d = {
            "class_module": self.__class__.__module__,
            "rule_class": self.__class__.__name__,
        }
        return d

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, d: dict) -> "AbstractRule":
        module = import_module(d.pop("class_module"))
        RuleClass: Type["AbstractRule"] = getattr(module, d.pop("rule_class"))
        if not issubclass(RuleClass, AbstractRule):
            raise ValueError("Not a valid rule class")
        return RuleClass.from_dict(d)

    @property
    def strategy(
        self,
    ) -> "AbstractStrategy[CombinatorialClassType, CombinatorialObjectType]":
        return self._strategy

    @property
    def ignore_parent(self) -> bool:
        """
        Return True if it is not worth expanding the parent/comb_class if
        this rule is found.
        """
        return self._strategy.ignore_parent

    @property
    def inferrable(self) -> bool:
        """
        Return True if the children could change using inferral strategies.
        """
        return self._strategy.inferrable

    @property
    def possibly_empty(self) -> bool:
        """
        Return True if it is possible that a child is empty.
        """
        return self._strategy.possibly_empty

    @property
    def workable(self) -> bool:
        """
        Return True if the children can expanded using other strategies.
        """
        return self._strategy.workable

    def set_subrecs(
        self,
        get_subrule: Callable[
            [CombinatorialClassType],
            "AbstractRule[CombinatorialClassType, CombinatorialObjectType]",
        ],
    ) -> None:
        """
        In general a rule comes in the context of a set of rules, e.g. in a
        CombinatorialSpecification. In order to count, generate, etc using the
        rule we must set the subfunctions that the children refer to.
        """
        self.subrecs = tuple(
            get_subrule(child).count_objects_of_size for child in self.children
        )
        self.subsamplers = tuple(
            get_subrule(child).random_sample_object_of_size for child in self.children
        )
        self.subobjects = tuple(
            get_subrule(child).get_objects for child in self.children
        )
        self.subterms = tuple(get_subrule(child).get_terms for child in self.children)

    @property
    def children(self) -> Tuple[CombinatorialClassType, ...]:
        """
        Return the tuple of CombinatorialClass that are found by applying the
        decomposition function of the strategy to the parent.
        """
        if self._children is None:
            self._children = self._strategy.decomposition_function(self.comb_class)
            if self._children is None:
                raise StrategyDoesNotApply(f"{self.strategy} does not apply")
        return self._children

    @property
    def formal_step(self) -> str:
        """
        Return a string with a short explanation about the rule.
        """
        return self._strategy.formal_step()

    @abc.abstractmethod
    def is_equivalence(self) -> bool:
        """
        Returns True if the rule is an equivalence.
        """

    def is_two_way(self) -> bool:
        """
        Returns True if it is a two way rule.
        """
        return self.strategy.is_two_way(self.comb_class)

    def is_reversible(self) -> bool:
        """
        Returns True if it is possible to compute the enumeration of a child using the
        enumeration of the parent and the other children.
        """
        return self.strategy.is_reversible(self.comb_class)

    def shifts(self) -> Tuple[int, ...]:
        if self._shifts is None:
            self._shifts = self.strategy.shifts(self.comb_class, self.children)
        return self._shifts

    @abc.abstractmethod
    def forest_key(
        self, get_label: Callable[[CombinatorialClassType], int]
    ) -> ForestRuleKey:
        """Return the forest key of the rule."""

    def non_empty_children(self) -> Tuple[CombinatorialClassType, ...]:
        """
        Return the tuple of non-empty combinatorial classes that are found
        by applying the decomposition function.
        """
        if self._non_empty_children is None:
            self._non_empty_children = tuple(
                child for child in self.children if not child.is_empty()
            )
        return self._non_empty_children

    @abc.abstractmethod
    def _ensure_level(self, n: int) -> None:
        """
        Ensures the terms are computed and in the terms_cache upto size n.
        """

    def get_terms(self, n: int) -> Terms:
        """
        Return the terms for the given n.
        """
        self._ensure_level(n)
        return self.terms_cache[n]

    def count_objects_of_size(self, n: int, **parameters: int) -> int:
        """
        The function count the objects with respect to the parameters. The
        result is cached.
        """
        terms = self.get_terms(n)
        params_tuple = tuple(parameters[k] for k in self.comb_class.extra_parameters)
        return terms[params_tuple]

    @abc.abstractmethod
    def get_equation(
        self, get_function: Callable[[CombinatorialClassType], Function]
    ) -> Eq:
        """
        Return the equation for the (ordinary) generating function.
        """

    @abc.abstractmethod
    def _ensure_level_objects(self, n: int) -> None:
        """
        Ensures the objects are computed and in the objects cache upto size n.
        """

    def get_objects(self, n: int) -> Objects:
        """
        Return the objects for the given n.
        """
        self._ensure_level_objects(n)
        return self.objects_cache[n]

    def generate_objects_of_size(
        self, n: int, **parameters: int
    ) -> Iterator[CombinatorialObjectType]:
        """
        Generate the objects by using the underlying bijection between the
        parent and children.
        """
        objects = self.get_objects(n)
        params_tuple = tuple(parameters[k] for k in self.comb_class.extra_parameters)
        yield from objects[params_tuple]

    @abc.abstractmethod
    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        """Return a random objects of the give size."""

    @abc.abstractmethod
    def sanity_check(self, n: int) -> bool:
        """
        Sanity check that this is a valid rule.

        Raise a SanityCheckFailure error if the sanity_check fails.
        """

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractRule):
            return NotImplemented
        return (
            isinstance(other, self.__class__)
            and self.comb_class == other.comb_class
            and self._strategy == other._strategy
        )

    def get_eq_symbol(self):
        return self.strategy.get_eq_symbol()

    def get_op_symbol(self):
        return self.strategy.get_op_symbol()

    def __str__(self) -> str:
        def frontpad(res, height):
            n = max(len(s) for s in res)
            for _ in range(height - len(res)):
                res.append(" " * n)

        def backpad(res):
            n = max(len(s) for s in res)
            for i, s in enumerate(res):
                res[i] = s + " " * (n - len(s))

        def join(res, new):
            backpad(new)
            backpad(res)
            frontpad(res, len(new))
            for i, s in enumerate(new):
                res[i] += s

        res = str(self.comb_class).split("\n")
        backpad(res)
        if self.children:
            children = [str(child).split("\n") for child in self.children]
            symbol_height = min(1, len(res) - 1)
            eq_symbol = (
                ["     " for i in range(symbol_height)]
                + [f"  {self.get_eq_symbol()}  "]
                + ["     " for i in range(symbol_height)]
            )
            join(res, eq_symbol)
            join(res, children[0])
            if len(children) > 1:
                op_symbol = (
                    ["     " for i in range(symbol_height)]
                    + [f"  {self.get_op_symbol()}  "]
                    + ["     " for i in range(symbol_height)]
                )
                for child in children[1:]:
                    join(res, op_symbol)
                    join(res, child)
        return f"{self.formal_step}\n" + "\n".join(x for x in res)


class Rule(AbstractRule[CombinatorialClassType, CombinatorialObjectType]):
    def __init__(
        self,
        strategy: "Strategy[CombinatorialClassType, CombinatorialObjectType]",
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ):
        super().__init__(strategy, comb_class, children)
        self._constructor: Optional[Constructor] = None

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["comb_class"] = self.comb_class.to_jsonable()
        d["children"] = [child.to_jsonable() for child in self.children]
        d["strategy"] = self.strategy.to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "Rule":
        # pylint: disable=import-outside-toplevel
        from comb_spec_searcher.strategies.strategy import AbstractStrategy, Strategy

        strategy = AbstractStrategy.from_dict(d.pop("strategy"))
        assert isinstance(strategy, Strategy)
        comb_class = CombinatorialClass.from_dict(d.pop("comb_class"))
        comb_class = cast(CombinatorialClassType, comb_class)
        d.pop("children")
        assert not d
        return cls(strategy, comb_class)

    @property
    def strategy(self) -> "Strategy[CombinatorialClassType, CombinatorialObjectType]":
        return cast(
            "Strategy[CombinatorialClassType, CombinatorialObjectType]",
            super().strategy,
        )

    @property
    def constructor(
        self,
    ) -> Constructor[CombinatorialClassType, CombinatorialObjectType]:
        """
        Return the constructor, that contains all the information about how to
        count/generate objects from the rule.
        """
        if self._constructor is None:
            self._constructor = self.strategy.constructor(
                self.comb_class, self.children
            )
            if self._constructor is None:
                raise StrategyDoesNotApply(f"{self.strategy} does not apply")
        return self._constructor

    def is_equivalence(self):
        return self.strategy.can_be_equivalent() and len(self.non_empty_children()) == 1

    def forest_key(
        self, get_label: Callable[[CombinatorialClassType], int]
    ) -> ForestRuleKey:
        return ForestRuleKey(
            get_label(self.comb_class),
            tuple(map(get_label, self.children)),
            self.shifts(),
            RuleBucket.EQUIV if self.is_equivalence() else RuleBucket.NORMAL,
        )

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        """
        This encodes the backward map of the underlying bijection that the
        strategy implies.
        """
        yield from self.strategy.backward_map(self.comb_class, objs, self.children)

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        """
        This encodes the forward map of the underlying bijection that the
        strategy implies.
        """
        return self.strategy.forward_map(self.comb_class, obj, self.children)

    def indexed_forward_map(
        self,
        obj: CombinatorialObjectType,
        data: Optional[object] = None,
    ) -> Tuple[Tuple[Optional[CombinatorialObjectType], ...], int]:
        """
        A version of forward_map that is used for bijections. The idx is used
        create a bijection from a nonbijective rule.
        """
        return self.forward_map(obj), 0

    def indexed_backward_map(
        self,
        objs: Tuple[Optional[CombinatorialObjectType], ...],
        idx: int,
        data: Optional[object] = None,
    ) -> CombinatorialObjectType:
        """
        A version of backward_map that is used for bijections. The idx is used
        create a bijection from a nonbijective rule.
        """
        return next(self.backward_map(objs))

    def to_equivalence_rule(self) -> "EquivalenceRule":
        """
        Return the version of the rule where the empty children are removed.
        """
        assert (
            self.is_equivalence()
        ), f"EquivalenceRule can only be created for equivalence rules\n{self}"
        return EquivalenceRule(self)

    def to_reverse_rule(self, idx: int) -> "Rule":
        """
        Return the reverse rule where the child at the given index is the parent.
        """
        assert self.is_reversible()
        return ReverseRule(self, idx)

    def _ensure_level(self, n: int) -> None:
        if self.subterms is None:
            raise RuntimeError("set_subrecs must be set first")
        while n >= len(self.terms_cache):
            terms = self.constructor.get_terms(
                self.get_terms, self.subterms, len(self.terms_cache)
            )
            self.terms_cache.append(terms)

    def get_equation(
        self, get_function: Callable[[CombinatorialClassType], Function]
    ) -> Eq:
        """
        Return the equation for the (ordinary) generating function.
        """
        lhs_func = get_function(self.comb_class)
        rhs_funcs = tuple(get_function(comb_class) for comb_class in self.children)
        return self.constructor.get_equation(lhs_func, rhs_funcs)

    def _ensure_level_objects(self, n: int) -> None:
        if self.subobjects is None:
            raise RuntimeError("set_subrecs must be set first")
        while n >= len(self.objects_cache):
            objects: Objects = defaultdict(list)
            for parameters, subobjects in self.constructor.get_sub_objects(
                self.subobjects, len(self.objects_cache)
            ):
                for sub_objs in product(*subobjects):
                    objects[parameters].extend(self.backward_map(sub_objs))
            self.objects_cache.append(objects)

    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        assert self.subrecs is not None, "you must call the set_subrecs function first"
        assert (
            self.subsamplers is not None
        ), "you must call the set_subrecs function first"
        total_count = self.count_objects_of_size(n=n, **parameters)
        subobjs = self.constructor.random_sample_sub_objects(
            total_count, self.subsamplers, self.subrecs, n, **parameters
        )
        objs = tuple(self.backward_map(subobjs))
        return random.choice(objs)

    def sanity_check(self, n: int) -> bool:
        try:
            return (
                self._sanity_check_count(n)
                and self._sanity_check_objects(n)
                and self._sanity_check_random_sample(n)
            )
        except SanityCheckFailure as e:
            raise e

    def _sanity_check_count(self, n: int) -> bool:
        """
        Sanity check that the count given by the rule matches the brute force count.
        """
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        actual_terms = self.comb_class.get_terms(n)
        temp_subterms = self.subterms
        self.subterms = tuple(child.get_terms for child in self.children)
        rule_terms = self.get_terms(n)
        self.subterms = temp_subterms
        if actual_terms != rule_terms:
            raise SanityCheckFailure(
                f"The following rule failed sanity check:\n"
                f"{self}\n"
                f"Failed for size {n}\n"
                f"The actual count is {actual_terms}.\n"
                f"The rule count is {rule_terms}.",
            )
        return True

    def _sanity_check_objects(self, n: int) -> bool:
        """
        Sanity check that the object given by the rule matches the brute force
        generated objects.
        """
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        tempobjects = self.subobjects
        self.subobjects = tuple(child.get_objects for child in self.children)
        try:
            rule_objects = self.get_objects(n)
        except NotImplementedError:
            # Skipping testing rules that have not implemented object generation.
            return True
        self.subobjects = tempobjects
        actual_objects = self.comb_class.get_objects(n)
        for obj_list in chain(rule_objects.values(), actual_objects.values()):
            obj_list.sort()
        if actual_objects != rule_objects:
            raise SanityCheckFailure(
                f"The following rule failed sanity check:\n"
                f"{self}\n"
                f"Failed for size {n}:\n"
                f"The actual objects are {actual_objects}\n"
                f"The rule generated objects {rule_objects}."
            )
        return True

    def _sanity_check_random_sample(self, n: int) -> bool:
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init
        actual_objects = self.comb_class.get_objects(n)
        possible_parameters = [
            (dict(zip(self.comb_class.extra_parameters, param)), param)
            for param in actual_objects
        ]
        if not possible_parameters:
            return True

        def fake_subsampler(
            comb_class: CombinatorialClassType,
        ) -> Callable[..., CombinatorialObjectType]:
            objs_cache: Dict[int, Objects] = {}

            def sampler(n: int, **parameters: int) -> CombinatorialObjectType:
                if n not in objs_cache:
                    objs_cache[n] = comb_class.get_objects(n)
                objs = objs_cache[n]
                param_tuple = tuple(parameters[p] for p in comb_class.extra_parameters)
                return cast(CombinatorialObjectType, random.choice(objs[param_tuple]))

            return sampler

        def fake_subrec(comb_class: CombinatorialClassType) -> Callable[..., int]:
            terms_cache: Dict[int, Terms] = {}

            def subrec(n: int, **parameters: int) -> int:
                if n not in terms_cache:
                    terms_cache[n] = comb_class.get_terms(n)
                terms = terms_cache[n]
                param_tuple = tuple(parameters[p] for p in comb_class.extra_parameters)
                return terms[param_tuple]

            return subrec

        tmpsamplers = self.subsamplers
        tmpsubrec = self.subrecs
        tmpsubterms = self.subterms
        self.subsamplers = tuple(fake_subsampler(child) for child in self.children)
        self.subrecs = tuple(fake_subrec(child) for child in self.children)
        self.subterms = tuple(child.get_terms for child in self.children)
        try:
            self.random_sample_object_of_size(n, **possible_parameters[0][0])
        except NotImplementedError:
            # Skipping testing rules that have not implemented random_sampling.
            self.subsamplers = tmpsamplers
            self.subrecs = tmpsubrec
            self.subterms = tmpsubterms
            return True

        for paramd, paramt in possible_parameters:
            obj = self.random_sample_object_of_size(n, **paramd)
            if obj not in actual_objects[paramt]:
                self.subsamplers = tmpsamplers
                self.subrecs = tmpsubrec
                self.subterms = tmpsubterms
                raise SanityCheckFailure(
                    f"The following rule failed sanity check:\n"
                    f"{self}\n"
                    f"Failed for size {n}:\n"
                    f"Sampled the object {obj} for the parameters {paramd}\n"
                )
        self.subsamplers = tmpsamplers
        self.subrecs = tmpsubrec
        self.subterms = tmpsubterms
        return True


class NonBijectiveRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    def indexed_forward_map(
        self,
        obj: CombinatorialObjectType,
        data: Optional[object] = None,
    ) -> Tuple[Tuple[Optional[CombinatorialObjectType], ...], int]:
        img = self.forward_map(obj)
        return img, self._forward_order(obj, img, data)

    def indexed_backward_map(
        self,
        objs: Tuple[Optional[CombinatorialObjectType], ...],
        idx: int,
        data: Optional[object] = None,
    ) -> CombinatorialObjectType:
        return self._backward_order_item(idx, objs, data)

    @abc.abstractmethod
    def _forward_order(
        self,
        obj: CombinatorialObjectType,
        image: Tuple[Optional[CombinatorialObjectType], ...],
        data: Optional[object] = None,
    ) -> int:
        """For rules that do not have an injective forward_map, we can mark the
        resulting object with a number i (that depends on the object we map from)
        in order to achieve a bijection. The backward map will then take said
        number and yield the i-th element (by some predefined order). This requires
        the backward_map's order to be deterministic.
        """

    @abc.abstractmethod
    def _backward_order_item(
        self,
        idx: int,
        objs: Tuple[Optional[CombinatorialObjectType], ...],
        data: Optional[object] = None,
    ) -> CombinatorialObjectType:
        """Given an iterator of all possible domain elements for an indexed forward map
        and an index, this method will return the element that maps to said index.
        """


class EquivalenceRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    def __init__(self, rule: Rule[CombinatorialClassType, CombinatorialObjectType]):
        non_empty_children = rule.non_empty_children()
        assert rule.is_equivalence(), f"not an equivalence rule: {rule}"
        child = non_empty_children[0]
        super().__init__(rule.strategy, rule.comb_class, (child,))
        self.child_idx = rule.children.index(child)
        self.actual_children = rule.children
        self.original_rule = rule
        self._constructor: Optional[Union[DisjointUnion, Complement]] = None

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("comb_class")
        d.pop("strategy")
        d.pop("children")
        d["original_rule"] = self.original_rule.to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "EquivalenceRule":
        rule = AbstractRule.from_dict(d.pop("original_rule"))
        assert isinstance(rule, Rule)
        assert not d, d.keys()
        return cls(rule)

    @property
    def constructor(self) -> Union[DisjointUnion, Complement]:
        if self._constructor is None:
            original_constructor = self.original_rule.constructor
            if isinstance(original_constructor, DisjointUnion):
                self._constructor = DisjointUnion(
                    self.comb_class,
                    self.children,
                    (original_constructor.extra_parameters[self.child_idx],),
                )
            elif isinstance(original_constructor, Complement):
                assert isinstance(self.original_rule, ReverseRule)
                original_original_rule = self.original_rule.original_rule
                original_original_child_idx = (
                    original_original_rule.to_equivalence_rule().child_idx
                )
                original_original_constructor = original_original_rule.constructor
                assert isinstance(original_original_constructor, DisjointUnion)
                self._constructor = Complement(
                    self.children[0],
                    (self.comb_class,),
                    0,
                    (
                        original_original_constructor.extra_parameters[
                            original_original_child_idx
                        ],
                    ),
                )
            else:
                raise NotImplementedError
        return self._constructor

    @staticmethod
    def is_equivalence() -> bool:
        return True

    def to_reverse_rule(self, idx: int) -> "EquivalenceRule":
        assert idx == 0
        return self.original_rule.to_reverse_rule(self.child_idx).to_equivalence_rule()

    def to_equivalence_rule(self) -> "EquivalenceRule":
        raise NotImplementedError("You don't want to do that! I promise")

    @property
    def formal_step(self) -> str:
        strat = self.strategy.formal_step()
        if isinstance(self.original_rule, ReverseRule):
            idx = self.original_rule.idx
        else:
            idx = self.child_idx
        s = f"{strat} but only child and index {idx} is non-empty"
        if isinstance(self.original_rule, ReverseRule):
            return f"reverse of '{s}'"
        return s

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        actual_objs = tuple(
            objs[0] if i == self.child_idx else None
            for i in range(len(self.actual_children))
        )
        yield from self.original_rule.backward_map(actual_objs)

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        return (self.original_rule.forward_map(obj)[self.child_idx],)


class EquivalencePathRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    """
    A class for shortening a chain of disjoint union  or complement unary rules into a
    single Rule.
    """

    def __init__(
        self, rules: Sequence[Rule[CombinatorialClassType, CombinatorialObjectType]]
    ):
        assert all(rule.is_equivalence() for rule in rules)
        assert all(len(rule.children) == 1 for rule in rules)
        super().__init__(rules[0].strategy, rules[0].comb_class, rules[-1].children)
        self.rules = tuple(rules)
        self._constructor: Optional[DisjointUnion] = None

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("comb_class")
        d.pop("strategy")
        d.pop("children")
        d["rules"] = [r.to_jsonable() for r in self.rules]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "EquivalencePathRule":
        rules: List[Rule] = []
        for rule_dict in d.pop("rules"):
            rule = AbstractRule.from_dict(rule_dict)
            assert isinstance(rule, Rule)
            rules.append(rule)
        assert not d
        return cls(rules)

    @property
    def constructor(self) -> DisjointUnion:
        if self._constructor is None:
            extra_parameters: Dict[str, str] = {
                k: k for k in self.comb_class.extra_parameters
            }
            for rule in self.rules:
                original_constructor = rule.constructor
                assert isinstance(original_constructor, (DisjointUnion, Complement))
                rules_parameters = original_constructor.extra_parameters[0]
                if isinstance(original_constructor, Complement):
                    rules_parameters = {b: a for a, b in rules_parameters.items()}
                extra_parameters = {
                    parent_var: rules_parameters[child_var]
                    for parent_var, child_var in extra_parameters.items()
                    if child_var in rules_parameters
                }
            fixed_values = {
                k: 0
                for k in self.children[0].extra_parameters
                if k not in extra_parameters.values()
            }
            self._constructor = DisjointUnion(
                self.comb_class, self.children, (extra_parameters,), (fixed_values,)
            )
        return self._constructor

    @staticmethod
    def is_equivalence() -> bool:
        return True

    @property
    def formal_step(self) -> str:
        return ", then ".join(rule.formal_step for rule in self.rules)

    def eqv_path_rules(self) -> List[Tuple[CombinatorialClassType, Rule]]:
        """
        Returns the list of (parent, rule) pairs that make up the equivalence
        path.
        """
        eqv_path_rules = []
        curr = self.comb_class
        for rule in self.rules:
            eqv_path_rules.append((curr, rule))
            curr = rule.children[0]
        return eqv_path_rules

    def to_equivalence_rule(self) -> "EquivalenceRule":
        raise NotImplementedError("You don't want to do that! I promise")

    def to_reverse_rule(self, idx: int) -> "Rule":
        raise NotImplementedError("You don't want to do that! I promise")

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        res = cast(Tuple[CombinatorialObjectType], objs)
        for rule in reversed(self.rules):
            try:
                res = (next(rule.backward_map(res)),)
            except StopIteration:
                return
        yield res[0]

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        res = cast(CombinatorialObjectType, obj)
        for rule in self.rules:
            res = cast(CombinatorialObjectType, rule.forward_map(res)[0])
        return (res,)

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and [r.strategy for r in self.rules] == [
            r.strategy for r in other.rules
        ]

    def __str__(self) -> str:
        def frontpad(res, height):
            n = max(len(s) for s in res)
            for _ in range(height - len(res)):
                res.append(" " * n)

        def backpad(res):
            n = max(len(s) for s in res)
            for i, s in enumerate(res):
                res[i] = s + " " * (n - len(s))

        def join(res, new):
            backpad(new)
            backpad(res)
            frontpad(res, len(new))
            for i, s in enumerate(new):
                res[i] += s

        res = str(self.comb_class).split("\n")
        backpad(res)
        comb_classes = [str(rule.children[0]).split("\n") for rule in self.rules]
        symbol_height = min(1, len(res) - 1)
        eq_symbol = (
            ["     " for i in range(symbol_height)]
            + ["  =  "]
            + ["     " for i in range(symbol_height)]
        )
        for comb_class in comb_classes:
            join(res, eq_symbol)
            join(res, comb_class)
        return f"{self.formal_step}\n" + "\n".join(x for x in res)


class ReverseRule(Rule[CombinatorialClassType, CombinatorialObjectType]):
    def __init__(
        self, rule: Rule[CombinatorialClassType, CombinatorialObjectType], idx: int
    ):
        assert rule.is_reversible()
        super().__init__(
            rule.strategy,
            rule.children[idx],
            (rule.comb_class, *rule.children[:idx], *rule.children[idx + 1 :]),
        )
        self.original_rule = rule
        self.idx = idx  # the idx of the child to be counted.
        self._constructor: Optional[Constructor] = None

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("comb_class")
        d.pop("strategy")
        d.pop("children")
        d["original_rule"] = self.original_rule.to_jsonable()
        d["idx"] = self.idx
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "ReverseRule":
        rule = AbstractRule.from_dict(d.pop("original_rule"))
        assert isinstance(rule, Rule)
        idx = d.pop("idx")
        assert not d
        return cls(rule, idx)

    def to_reverse_rule(self, idx: int) -> "Rule":
        raise NotImplementedError("You don't want to do that! I promise")

    def shifts(self) -> Tuple[int, ...]:
        original_shifts = self.original_rule.shifts()
        pshift = -original_shifts[self.idx]
        return (pshift,) + tuple(
            s + pshift
            for s in (s for i, s in enumerate(original_shifts) if i != self.idx)
        )

    def forest_key(
        self, get_label: Callable[[CombinatorialClassType], int]
    ) -> ForestRuleKey:
        return ForestRuleKey(
            get_label(self.comb_class),
            tuple(map(get_label, self.children)),
            self.shifts(),
            RuleBucket.EQUIV if self.is_equivalence() else RuleBucket.REVERSE,
        )

    def get_equation(
        self, get_function: Callable[[CombinatorialClassType], Function]
    ) -> Eq:
        try:
            return super().get_equation(get_function)
        except NotImplementedError:
            return self.original_rule.get_equation(get_function)

    @property
    def constructor(
        self,
    ) -> Constructor:
        if self._constructor is None:
            self._constructor = self.strategy.reverse_constructor(
                self.idx,
                self.original_rule.comb_class,
                self.original_rule.children,
            )
        return self._constructor

    @property
    def formal_step(self) -> str:
        if len(self.children) == 1:
            return f"reverse of '{self.strategy.formal_step()}'"
        return f"reverse of '{self.strategy.formal_step()}', counting child {self.idx}"

    def backward_map(
        self, objs: Tuple[Optional[CombinatorialObjectType], ...]
    ) -> Iterator[CombinatorialObjectType]:
        if not len(self.original_rule.non_empty_children()) == 1:
            raise NotImplementedError("Cannot map forward for non equivalence rule.")
        assert all(obj is None for obj in objs[1:])
        assert objs[0] is not None
        res = self.original_rule.forward_map(objs[0])[self.idx]
        assert res is not None
        yield res

    def forward_map(
        self, obj: CombinatorialObjectType
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        if not len(self.original_rule.non_empty_children()) == 1:
            raise NotImplementedError("Cannot map forward for non equivalence rule.")
        objs: List[Optional[CombinatorialObjectType]] = [
            None for _ in self.original_rule.children
        ]
        objs[self.idx] = obj
        orig_res = list(self.original_rule.backward_map(tuple(objs)))
        assert len(orig_res) == 1, "More than one results from back map"
        return tuple(orig_res) + tuple(None for _ in range(len(self.children) - 1))

    def get_op_symbol(self):
        return self.strategy.get_reverse_op_symbol()


class VerificationRule(AbstractRule[CombinatorialClassType, CombinatorialObjectType]):
    """
    A class for verification rules. In particular, the children will be an
    empty tuple if it applies, else None.
    """

    # The signature is refined from AbstractRule
    # pylint: disable=useless-super-delegation
    def __init__(
        self,
        strat: "VerificationStrategy[CombinatorialClassType,CombinatorialObjectType]",
        comb_class: CombinatorialClassType,
        children: Optional[Tuple[CombinatorialClassType, ...]] = None,
    ):
        super().__init__(strat, comb_class, children)

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["comb_class"] = self.comb_class.to_jsonable()
        d["strategy"] = self.strategy.to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "VerificationRule":
        # pylint: disable=import-outside-toplevel
        from comb_spec_searcher.strategies.strategy import (
            AbstractStrategy,
            VerificationStrategy,
        )

        strategy = AbstractStrategy.from_dict(d.pop("strategy"))
        assert isinstance(strategy, VerificationStrategy)
        comb_class = CombinatorialClass.from_dict(d.pop("comb_class"))
        comb_class = cast(CombinatorialClassType, comb_class)
        assert not d
        return cls(strategy, comb_class)

    @property
    def strategy(
        self,
    ) -> "VerificationStrategy[CombinatorialClassType, CombinatorialObjectType]":
        return cast(
            "VerificationStrategy[CombinatorialClassType, CombinatorialObjectType]",
            super().strategy,
        )

    @staticmethod
    def is_equivalence() -> bool:
        return False

    def _ensure_level(self, n: int) -> None:
        while n >= len(self.terms_cache):
            terms = self.strategy.get_terms(self.comb_class, len(self.terms_cache))
            self.terms_cache.append(terms)

    def _ensure_level_objects(self, n: int) -> None:
        while n >= len(self.objects_cache):
            objects = self.strategy.get_objects(
                self.comb_class, len(self.objects_cache)
            )
            self.objects_cache.append(objects)

    def pack(self) -> "StrategyPack":
        return self.strategy.pack(self.comb_class)

    def forest_key(
        self, get_label: Callable[[CombinatorialClassType], int]
    ) -> ForestRuleKey:
        return ForestRuleKey(
            get_label(self.comb_class),
            tuple(map(get_label, self.children)),
            self.shifts(),
            RuleBucket.VERIFICATION,
        )

    # pylint: disable=arguments-differ
    def get_equation(
        self,
        get_function: Callable[[CombinatorialClassType], Function],
        funcs: Optional[Dict[CombinatorialClassType, Function]] = None,
    ) -> Eq:
        lhs_func = get_function(self.comb_class)
        return Eq(lhs_func, self.strategy.get_genf(self.comb_class, funcs))

    @staticmethod
    def get_eq_symbol() -> str:
        return "↝"

    def random_sample_object_of_size(
        self, n: int, **parameters: int
    ) -> CombinatorialObjectType:
        return self.strategy.random_sample_object_of_size(
            self.comb_class, n, **parameters
        )

    def sanity_check(self, n: int) -> bool:
        return self.get_terms(n) == self.comb_class.get_terms(n)



import random
from collections import defaultdict
from functools import partial
from itertools import product
from typing import (
    Callable,
    Counter,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import sympy

from comb_spec_searcher import utils
from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)

from .base import Constructor

T = TypeVar("T")


class CartesianProduct(Constructor[CombinatorialClassType, CombinatorialObjectType]):
    """
    The CartesianProduct is initialised with the children of the rule that is
    being counted. These are needed in the reliance profile. In particular,
    the CombinatorialClass that you are counting must have implemented the
    methods is_atom', 'minimum_size_of_object', 'get_minimum_values' which are
    needed to ensure that the recursions are productive.

    This CartesianProduct constructor considers compositions all of the
    parameters including n. then these should be passed to one other factor.
    The details of how the parameters map forward must be given using
    'extra_parameters'. This is a tuple, where the ith dictionary tells the
    constructor for each parent variable maps to on the child. If it does not
    map to the child it should not be a key in the dictionary.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Iterable[CombinatorialClassType],
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):

        children = tuple(children)

        if extra_parameters is not None:
            self.extra_parameters = tuple(extra_parameters)
        else:
            self.extra_parameters = tuple({} for _ in children)
        self._children_param_maps = self._build_children_param_map(
            parent, children, self.extra_parameters
        )

        self.minimum_sizes = {"n": parent.minimum_size_of_object()}
        for k in parent.extra_parameters:
            self.minimum_sizes[k] = parent.get_minimum_value(k)

        self.min_child_sizes = tuple(
            {"n": child.minimum_size_of_object()} for child in children
        )
        self.max_child_sizes = tuple(
            {"n": child.minimum_size_of_object()} if child.is_atom() else {}
            for child in children
        )
        self.parent_parameters = ("n",) + parent.extra_parameters

        for (idx, child), parameters in zip(enumerate(children), self.extra_parameters):
            for k in parent.extra_parameters:
                self.min_child_sizes[idx][k] = (
                    child.get_minimum_value(parameters[k]) if k in parameters else 0
                )
                if k not in parameters:
                    self.max_child_sizes[idx][k] = 0
                elif child.is_atom():
                    self.max_child_sizes[idx][k] = child.get_minimum_value(
                        parameters[k]
                    )

    @property
    def min_sizes(self) -> Tuple[int, ...]:
        """
        Lower bound on the sizes of combinatorial object on the children.
        """
        return tuple(d["n"] for d in self.min_child_sizes)

    @property
    def max_sizes(self) -> Tuple[Optional[int], ...]:
        """
        Upper bounds on the size of object on the children.

        None means no upper bound.
        """
        return tuple(d.get("n", None) for d in self.max_child_sizes)

    @staticmethod
    def _build_children_param_map(
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        extra_parameters: Tuple[Dict[str, str], ...],
    ) -> Tuple[ParametersMap, ...]:
        map_list: List[ParametersMap] = []
        num_parent_params = len(parent.extra_parameters)
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        for child, extra_param in zip(children, extra_parameters):
            reversed_extra_param: Dict[str, List[str]] = defaultdict(list)
            for parent_var, child_var in extra_param.items():
                reversed_extra_param[child_var].append(parent_var)
            child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
                tuple(
                    map(
                        parent_param_to_pos.__getitem__,
                        reversed_extra_param[child_param],
                    )
                )
                for child_param in child.extra_parameters
            )
            map_list.append(
                Constructor.build_param_map(child_pos_to_parent_pos, num_parent_params)
            )
        return tuple(map_list)

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        res = 1
        for extra_parameters, rhs_func in zip(self.extra_parameters, rhs_funcs):
            res *= rhs_func.subs(
                {child: parent for parent, child in extra_parameters.items()},
                simultaneous=True,
            )
        return sympy.Eq(lhs_func, res)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        assert all(
            set(["n", *parameters]) == set(min_child_sizes)
            for min_child_sizes in self.min_child_sizes
        )
        parameters["n"] = n
        return tuple(
            {
                k: tuple(
                    range(
                        min_child_sizes[k],
                        min(
                            [
                                parameters[k]
                                - self.minimum_sizes[k]
                                + min_child_sizes[k]
                                + 1
                            ]
                            + (
                                [max_child_sizes[k] + 1] if k in max_child_sizes else []
                            ),
                        ),
                    )
                )
                for k in min_child_sizes
            }
            for min_child_sizes, max_child_sizes in zip(
                self.min_child_sizes, self.max_child_sizes
            )
        )

    def _valid_compositions(
        self, n: int, **parameters: int
    ) -> Iterator[Tuple[Dict[str, int], ...]]:
        reliance_profile = self.reliance_profile(n, **parameters)

        def _helper(
            minmaxes: Tuple[Dict[str, Tuple[int, int]], ...], **parameters: int
        ):
            if len(minmaxes) == 1:
                minmax = minmaxes[0]
                if all(
                    minmax[k][0] <= parameters[k] <= minmax[k][1]
                    for k in self.parent_parameters
                ):
                    yield ({**parameters},)
                return

            still_to_come = {
                k: sum(minmax[k][0] for minmax in minmaxes[1:])
                for k in self.parent_parameters
            }
            max_available = {
                k: sum(minmax[k][1] for minmax in minmaxes[1:])
                for k in self.parent_parameters
            }
            minmax = minmaxes[0]
            for values in product(
                *[
                    range(
                        max(minmax[k][0], parameters[k] - max_available[k]),
                        min(minmax[k][1], parameters[k] - still_to_come[k]) + 1,
                    )
                    for k in self.parent_parameters
                ]
            ):
                params = dict(zip(self.parent_parameters, values))
                update_params = {
                    k: parameters[k] - params[k] for k in self.parent_parameters
                }
                for comp in _helper(minmaxes[1:], **update_params):
                    yield (params,) + comp

        if all(all(profile.values()) for profile in reliance_profile):
            minmaxes: Tuple[Dict[str, Tuple[int, int]], ...] = tuple(
                {k: (min(profile[k]), max(profile[k])) for k in self.parent_parameters}
                for profile in reliance_profile
            )
            parameters["n"] = n
            yield from _helper(minmaxes, **parameters)

    def get_extra_parameters(
        self, child_parameters: Tuple[Dict[str, int], ...]
    ) -> Optional[List[Dict[str, int]]]:
        """
        Will return the extra parameters dictionary based on the given child
        parameters given. If there is a contradiction, that is some child
        parameter is given two or more different values that do not match,
        then None will be returned indicating that there is a contradiction,
        and so there are no objects satisfying the child parameters.
        """
        res: List[Dict[str, int]] = []
        for params, map_params in zip(child_parameters, self.extra_parameters):
            assert all(
                params[k] == 0
                for k in self.parent_parameters
                if k not in map_params and k != "n"
            )
            extra_params: Dict[str, int] = {"n": params["n"]}
            for k in map_params:
                mapped_k = map_params[k]
                if mapped_k not in extra_params:
                    extra_params[mapped_k] = params[k]
                elif extra_params[mapped_k] != params[k]:
                    return None
            res.append(extra_params)
        return res

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        new_terms: Terms = Counter()
        size_compositions = utils.compositions(
            n, len(subterms), self.min_sizes, self.max_sizes
        )
        for sizes in size_compositions:
            for param_value_pairs in self.params_value_pairs_combinations(
                sizes, subterms
            ):
                new_param = self._new_param(*(p for p, _ in param_value_pairs))
                new_terms[new_param] += utils.prod((v for _, v in param_value_pairs))
        return new_terms

    def _new_param(self, *children_params: Parameters) -> Parameters:
        """
        Computes the parameter values on the parent that the given children parameters
        contribute to.
        """
        mapped_params = (
            pmap(p) for pmap, p in zip(self._children_param_maps, children_params)
        )
        return tuple(sum(vals) for vals in zip(*mapped_params))

    @staticmethod
    def params_value_pairs_combinations(
        sizes: Tuple[int, ...],
        sub_getters: Tuple[Callable[[int], Dict[Parameters, T]], ...],
    ) -> Iterator[Tuple[Tuple[Parameters, T], ...]]:
        assert len(sizes) == len(sub_getters), sub_getters
        children_values = (sg(s) for sg, s in zip(sub_getters, sizes))
        return product(*(c.items() for c in children_values))

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        size_compositions = utils.compositions(
            n, len(subobjs), self.min_sizes, self.max_sizes
        )
        for sizes in size_compositions:
            for param_objs_pairs in self.params_value_pairs_combinations(
                sizes, subobjs
            ):
                new_param = self._new_param(*(p for p, _ in param_objs_pairs))
                children_objs = cast(
                    Iterator[List[Optional[CombinatorialObjectType]]],
                    (objs for _, objs in param_objs_pairs),
                )
                yield (new_param, tuple(children_objs))

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        random_choice = random.randint(1, parent_count)
        total = 0
        for child_parameters in self._valid_compositions(n, **parameters):
            tmp = 1
            extra_parameters = self.get_extra_parameters(child_parameters)
            if extra_parameters is None:
                continue
            for rec, extra_params in zip(subrecs, extra_parameters):
                tmp *= rec(n=extra_params.pop("n"), **extra_params)
                if tmp == 0:
                    break
            total += tmp
            if random_choice <= total:
                extra_parameters = self.get_extra_parameters(child_parameters)
                assert extra_parameters is not None
                return tuple(
                    subsampler(n=extra_params.pop("n"), **extra_params)
                    for subsampler, extra_params in zip(subsamplers, extra_parameters)
                )

    @staticmethod
    def get_eq_symbol() -> str:
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        return "x"

    def __str__(self) -> str:
        return "Cartesian product"

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and CartesianProduct.extra_params_equiv(
                self.extra_parameters, other.extra_parameters
            ),
            None,
        )


class Quotient(Constructor[CombinatorialClassType, CombinatorialObjectType]):
    """
    This constructor implements the quotient operation. A rule A -> (B, C, D)
    should be read as A = B / (C * D).

    This comes from a cartesian product rule B = A * C * D

    The initialiser should be made using the original cartesian product rule.

    The parent is B, idx is the index of the child the rule is being rearranged
    for and the extra parameters maps from parent to children.

    To get terms we compute a polynomial equation for a fixed size a_n(x,...) =
    b_n(x,...) * c(x, ...) where b_n is the terms for the flipped child of length n and
    then make the division.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        idx: int,
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):
        self.number_of_children = len(children)
        if extra_parameters is not None:
            self.extra_parameters = tuple(extra_parameters)
        else:
            self.extra_parameters = tuple({} for _ in children)
        self.idx = idx
        self._children_param_maps = CartesianProduct._build_children_param_map(
            parent, children, self.extra_parameters
        )
        self._min_sizes = tuple(child.minimum_size_of_object() for child in children)
        self._max_sizes = tuple(
            child.minimum_size_of_object() if child.is_atom() else None
            for child in children
        )
        self._parent_param_map = self._build_parent_param_map(parent, children)
        self._parent_shift = sum(self._min_sizes) - self._min_sizes[self.idx]
        self._num_parent_params = len(parent.extra_parameters)

    @staticmethod
    def param_map(
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...],
        num_parent_params: int,
        param: Parameters,
    ) -> Parameters:
        new_params: List[Optional[int]] = [None for _ in range(num_parent_params)]
        for pos, value in enumerate(param):
            parent_pos = child_pos_to_parent_pos[pos]
            for p in parent_pos:
                assert new_params[p] is None or new_params[p] == value
                new_params[p] = value
        assert all(p is not None for p in new_params)
        return cast(Parameters, tuple(new_params))

    @staticmethod
    def build_param_map(
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...], num_parent_params: int
    ) -> ParametersMap:

        return partial(Quotient.param_map, child_pos_to_parent_pos, num_parent_params)

    def _build_parent_param_map(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
    ) -> ParametersMap:
        extra_param = self.extra_parameters[self.idx]
        child = children[self.idx]
        child_param_to_pos = {
            param: pos for pos, param in enumerate(child.extra_parameters)
        }
        parent_pos_to_child_pos: Tuple[Tuple[int, ...], ...] = tuple(
            (child_param_to_pos[extra_param[parent_var]],)
            if parent_var in extra_param
            else tuple()
            for parent_var in parent.extra_parameters
        )
        return self.build_param_map(
            parent_pos_to_child_pos, len(child.extra_parameters)
        )

    def _new_param(self, *children_params: Parameters) -> Parameters:
        """
        Computes the parameter values on the parent that the given children parameters
        contribute to.
        """
        mapped_params = (
            pmap(p) for pmap, p in zip(self._children_param_maps, children_params)
        )
        return tuple(sum(vals) for vals in zip(*mapped_params))

    def _other_new_param(self, *children_params: Parameters) -> Parameters:
        """
        Computes the parameter values on the parent that the given children parameters
        contribute to.
        """
        mapped_params = (
            pmap(p)
            for pmap, p in zip(
                (m for i, m in enumerate(self._children_param_maps) if i != self.idx),
                children_params,
            )
        )
        return tuple(sum(vals) for vals in zip(*mapped_params))

    def _a(
        self,
        n: int,
        parent_subterm: Callable[[int], Terms],
        children_subterms: SubTerms,
    ) -> Terms:
        """
        A terms dict representing the polynomial for the rule parent terms minus the
        contribution of the childs when the size of the flipped child in the composition
        is smaller than n.
        """
        max_sizes = (
            self._max_sizes[: self.idx] + (n - 1,) + self._max_sizes[self.idx + 1 :]
        )
        possible_sizes = utils.compositions(
            n + self._parent_shift, self.number_of_children, self._min_sizes, max_sizes
        )
        res: Terms = Counter(parent_subterm(n + self._parent_shift))
        for sizes in possible_sizes:
            for param_value_pairs in CartesianProduct.params_value_pairs_combinations(
                sizes, children_subterms
            ):
                new_param = self._new_param(*(p for p, _ in param_value_pairs))
                res[new_param] -= utils.prod((v for _, v in param_value_pairs))
                assert res[new_param] >= 0
        return res

    def _c(
        self,
        children_subterms: SubTerms,
    ) -> Terms:
        """
        Dict representing the polynomial that multiply the polynomial of the
        flipped child for length n.
        """
        min_sizes = self._min_sizes[: self.idx] + self._min_sizes[self.idx + 1 :]
        max_sizes = self._max_sizes[: self.idx] + self._max_sizes[self.idx + 1 :]
        possible_sizes = utils.compositions(
            self._parent_shift, self.number_of_children - 1, min_sizes, max_sizes
        )
        other_children_subterms = (
            children_subterms[: self.idx] + children_subterms[self.idx + 1 :]
        )
        res: Terms = Counter()
        for sizes in possible_sizes:
            for param_value_pairs in CartesianProduct.params_value_pairs_combinations(
                sizes, other_children_subterms
            ):
                new_param = self._other_new_param(*(p for p, _ in param_value_pairs))
                res[new_param] += utils.prod((v for _, v in param_value_pairs))
                assert res[new_param] >= 0
        return res

    def _terms_to_poly(self, terms: Terms) -> Union[sympy.Poly, int]:
        variables = tuple(sympy.var(f"k_{i}") for i in range(self._num_parent_params))
        if variables:
            res = sympy.Poly(sympy.sympify("0"), *variables)
        else:
            res = 0
        for param, value in terms.items():
            if not variables:
                res += value
            else:
                res += value * utils.prod(
                    var ** power for var, power in zip(variables, param)
                )
        return res

    @staticmethod
    def _poly_to_terms(poly: Union[sympy.Poly, int]) -> Terms:
        if isinstance(poly, (int, sympy.core.numbers.Integer)):
            if poly == 0:
                return Counter()
            return Counter({tuple(): int(poly)})
        return Counter(poly.as_dict())

    def _b(
        self,
        n: int,
        parent_subterm: Callable[[int], Terms],
        children_subterms: SubTerms,
    ) -> Terms:
        """
        The terms of length n for the flipped child.
        """
        a = self._a(n, parent_subterm, children_subterms)
        c = self._c(children_subterms)
        a_poly = self._terms_to_poly(a)
        c_poly = self._terms_to_poly(c)
        if self._num_parent_params > 0:
            b_poly, remainder = sympy.div(a_poly, c_poly, domain="ZZ")
            assert remainder == 0
        else:
            b_poly = a_poly // c_poly
            assert a_poly % c_poly == 0
        return self._poly_to_terms(b_poly)

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        new_terms: Terms = Counter()
        children_subterms = (
            subterms[1 : self.idx + 1] + (parent_terms,) + subterms[self.idx + 1 :]
        )
        for param, value in self._b(n, subterms[0], children_subterms).items():
            mapped_params = self._parent_param_map(param)
            assert mapped_params not in new_terms or new_terms[mapped_params] == value
            new_terms[mapped_params] = value
        return new_terms

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        if any(self.extra_parameters):
            raise NotImplementedError(
                "Quotient equation is not implemented with extra parameters. "
                "You can fall back on the cartesian equation."
            )
        res = rhs_funcs[0]
        for rhs_func in rhs_funcs[1:]:
            res /= rhs_func
        return sympy.Eq(lhs_func, res)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        raise NotImplementedError

    def __str__(self):
        return "quotient"

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and Quotient.extra_params_equiv(
                self.extra_parameters, other.extra_parameters
            ),
            None,
        )



import itertools
import time
from datetime import timedelta
from typing import (
    Callable,
    Deque,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from logzero import logger

from comb_spec_searcher.class_db import ClassDB
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.rule_db.abstract import RuleDBAbstract, ensure_specification
from comb_spec_searcher.strategies.rule import (
    AbstractRule,
    EquivalencePathRule,
    EquivalenceRule,
    Rule,
)
from comb_spec_searcher.strategies.strategy import (
    AbstractStrategy,
    EmptyStrategy,
    StrategyFactory,
)
from comb_spec_searcher.strategies.strategy_pack import StrategyPack
from comb_spec_searcher.typing import CSSstrategy, ForestRuleKey, RuleBucket, RuleKey

T = TypeVar("T")
RuleWithShifts = Tuple[RuleKey, Tuple[int, ...]]
SortedRWS = Dict[RuleBucket, List[ForestRuleKey]]
empty_strategy: EmptyStrategy = EmptyStrategy()


class DefaultList(Generic[T]):
    """
    A list data structure get automatically gets longer if an index not existing is
    requested.

    When getting longer the list is filled by calling the provied `default_factory`.
    This is similar to the `collections.defaultdict` data structure.
    """

    def __init__(self, default_factory: Callable[[], T]):
        self._default_factory = default_factory
        self._list: List[T] = []

    def _increase_list_len(self, key: int) -> None:
        """
        Increase the length of the list so that the given is valid.
        """
        num_new_entry = key - len(self._list) + 1
        self._list.extend((self._default_factory() for _ in range(num_new_entry)))

    def __getitem__(self, key: int) -> T:
        try:
            return self._list[key]
        except IndexError:
            self._increase_list_len(key)
            return self._list[key]

    def __setitem__(self, key: int, value: T) -> None:
        self._list[key] = value

    def __iter__(self) -> Iterator[T]:
        return iter(self._list)

    def __str__(self) -> str:
        return str(self._list)


class Function:
    """
    A python representation of a function.

    The function maps natural number to a natural number or infinity (represented
    by the use of None)

    The default value of the function is 0.
    """

    def __init__(self) -> None:
        self._value: List[Optional[int]] = []
        self._preimage_count: DefaultList[int] = DefaultList(int)
        self._infinity_count: int = 0

    @property
    def preimage_count(self) -> List[int]:
        """
        Return the number of classes for each value of the function.
        """
        count = list(self._preimage_count)
        while count and count[-1] == 0:
            count.pop()
        return count

    @property
    def infinity_count(self) -> int:
        return self._infinity_count

    def __getitem__(self, key: int) -> Optional[int]:
        """
        Return the value of the function for the given key.
        """
        try:
            return self._value[key]
        except IndexError:
            self._increase_list_len(key)
            return 0

    def _increase_list_len(self, key: int) -> None:
        num_new_entry = key - len(self._value) + 1
        self._value.extend((0 for _ in range(num_new_entry)))
        self._preimage_count[0] += num_new_entry

    def increase_value(self, key: int) -> None:
        """
        Increase by one the value of the function for the given key.
        """
        try:
            old_value = self._value[key]
            if old_value is None:
                raise ValueError(f"The function is already infinity for {key}")
        except IndexError:
            self._increase_list_len(key)
            old_value = 0
        self._value[key] = old_value + 1
        self._preimage_count[old_value] -= 1
        self._preimage_count[old_value + 1] += 1

    def set_infinite(self, key: int) -> None:
        """
        Set the value of function for the given key to infinity.
        """
        try:
            old_value = self._value[key]
            if old_value is None:
                raise ValueError(f"The function is already infinity for {key}")
        except IndexError:
            self._increase_list_len(key)
            old_value = 0
        self._value[key] = None
        self._preimage_count[old_value] -= 1
        self._infinity_count += 1

    def preimage_gap(self, length: int) -> int:
        """
        Return the smallest k such that the preimage of the interval
        [k, k+length-1] is empty.
        """
        if length <= 0:
            raise ValueError("length argument must be positive")
        last_non_zero = -1
        for i, v in enumerate(self._preimage_count):
            if v != 0:
                last_non_zero = i
            elif i - last_non_zero >= length:
                return last_non_zero + 1
        return last_non_zero + 1

    def preimage(self, value: Optional[int]) -> Iterator[int]:
        """
        Return the preimage of the function for the given value.
        """
        if value == 0:
            raise ValueError("The preimage of 0 is infinite.")
        return (k for k, v in enumerate(self._value) if v == value)

    def to_dict(self) -> Dict[int, Optional[int]]:
        """
        Return a dictionary view of the function with only the non-zero value.
        """
        return {i: v for i, v in enumerate(self._value) if v != 0}

    def __str__(self) -> str:
        parts = (
            f"{i} -> {v if v is not None else '∞'}" for i, v in enumerate(self._value)
        )
        return "\n".join(parts)


class TableMethod:
    def __init__(self) -> None:
        self._rules: List[ForestRuleKey] = []
        self._shifts: List[List[Optional[int]]] = []
        self._function: Function = Function()
        self._gap_size: int = 1
        self._rules_using_class: DefaultList[List[Tuple[int, int]]] = DefaultList(list)
        self._rules_pumping_class: DefaultList[List[int]] = DefaultList(list)
        self._processing_queue: Deque[int] = Deque()
        self._current_gap: Tuple[int, int] = (1, 1)
        self._rule_holding_extra_terms: Set[int] = set()

    @property
    def function(self) -> Dict[int, Optional[int]]:
        """
        Return a dict representing the number of term that are computable for each
        class.
        Only the class where it can get at least a term are included. If a class is map
        to None, then all the terms of the enumeration are computable.
        """
        return self._function.to_dict()

    def add_rule_key(
        self,
        rule_key: ForestRuleKey,
    ):
        """
        Add the rule to the database.

        INPUTS:
          - `rule_key`
          - `shifts_for_zero`: The values of the shifts if no information was known
          about any of the classes.
          - `rule_bucket` the type of rule
        """
        self._rules.append(rule_key)
        self._shifts.append(self._compute_shift(rule_key.key, rule_key.shifts))
        max_gap = max((abs(s) for s in rule_key.shifts), default=0)
        if max_gap > self._gap_size:
            self._gap_size = max_gap
            self._correct_gap()
        if self._function[rule_key.parent] is not None:
            rule_idx = len(self._rules) - 1
            self._rules_pumping_class[rule_key.parent].append(rule_idx)
            for child_idx, child in enumerate(rule_key.children):
                if self._function[child] is not None:
                    self._rules_using_class[child].append((rule_idx, child_idx))
            self._processing_queue.append(rule_idx)
        self._process_queue()

    def is_pumping(self, label: int) -> bool:
        """
        Determine if the comb_class is pumping in the current universe.
        """
        return self._function[label] is None

    def status(self) -> str:
        s = f"\tSize of the gap: {self._gap_size}\n"
        s += f"\tSize of the stable subset: {self._function.infinity_count}\n"
        s += f"\tSizes of the pre-images: {self._function.preimage_count}\n"
        return s

    def stable_subset(self) -> Iterator[int]:
        return self._function.preimage(None)

    def pumping_subuniverse(
        self,
    ) -> Iterator[ForestRuleKey]:
        """
        Iterator over all the forest rule keys that contain only pumping
        combinatorial classes.
        """
        stable_subset = set(self.stable_subset())
        for forest_key in self._rules:
            if forest_key.parent in stable_subset and stable_subset.issuperset(
                forest_key.children
            ):
                yield forest_key

    def _compute_shift(
        self,
        rule_key: RuleKey,
        shifts_for_zero: Tuple[int, ...],
    ) -> List[Optional[int]]:
        """
        Compute the initial value for the shifts a rule based on the current state of
        the function.
        """
        parent_current_value = self._function[rule_key[0]]
        if parent_current_value is None:
            return [None for _ in shifts_for_zero]
        chidlren_function_value = map(self._function.__getitem__, rule_key[1])
        return [
            fvalue + sfz - parent_current_value if fvalue is not None else None
            for fvalue, sfz in zip(chidlren_function_value, shifts_for_zero)
        ]

    def _correct_gap(self) -> None:
        """
        Correct the gap and if needed queue rules for the classes that were previously
        on the right hand  side of the gap.

        This should be toggled every time the gap changes whether the size changes or
        the some value changes of the function caused the gap to change.
        """
        k = self._function.preimage_gap(self._gap_size)
        new_gap = (k, k + self._gap_size - 1)
        if new_gap[1] > self._current_gap[1]:
            self._processing_queue.extend(self._rule_holding_extra_terms)
            self._rule_holding_extra_terms.clear()
        self._current_gap = new_gap

    def _process_queue(self) -> None:
        """
        Try to make improvement with all the class in the processing queue.
        """
        while self._processing_queue or self._rule_holding_extra_terms:
            while self._processing_queue:
                rule_idx = self._processing_queue.popleft()
                shifts = self._shifts[rule_idx]
                if self._can_give_terms(shifts):
                    parent = self._rules[rule_idx].parent
                    self._increase_value(parent, rule_idx)
            if self._rule_holding_extra_terms:
                rule_idx = self._rule_holding_extra_terms.pop()
                parent = self._rules[rule_idx].parent
                self._set_infinite(parent)

    @staticmethod
    def _can_give_terms(shifts: List[Optional[int]]) -> bool:
        """
        Return True if the shifts indicate that a new terms can be computed.
        """
        return all(s is None or s > 0 for s in shifts)

    def _increase_value(self, comb_class: int, rule_idx: int) -> None:
        """
        Increase the value of the comb_class and put on the processing stack any rule
        that can now give a new term.

        The rule_idx must indicate the rule used to justify the increase.
        """
        current_value = self._function[comb_class]
        if current_value is None:
            return
        if current_value > self._current_gap[1]:
            self._rule_holding_extra_terms.add(rule_idx)
            return
        self._function.increase_value(comb_class)
        # Correction of the gap
        gap_start = self._function.preimage_gap(self._gap_size)
        if self._current_gap[0] != gap_start:
            self._correct_gap()
        # Correction of the shifts for rule pumping comb_class
        for r_idx in self._rules_pumping_class[comb_class]:
            shifts = self._shifts[r_idx]
            for i, v in enumerate(shifts):
                shifts[i] = v - 1 if v is not None else None
            if self._can_give_terms(shifts):
                self._processing_queue.append(r_idx)
        # Correction of the shifts for rules using comb_class to pump
        for r_idx, class_idx in self._rules_using_class[comb_class]:
            shifts = self._shifts[r_idx]
            current_shift = shifts[class_idx]
            assert current_shift is not None
            shifts[class_idx] = current_shift + 1
            if self._can_give_terms(shifts):
                self._processing_queue.append(r_idx)

    def _set_infinite(self, comb_class: int) -> None:
        """
        Set the value if the class to infinity.

        This should happen when we know that we cannot pump anything on the left side
        of the gap.
        """
        current_value = self._function[comb_class]
        if current_value is None:
            return
        assert current_value > self._current_gap[1]
        assert not self._processing_queue
        self._function.set_infinite(comb_class)
        # This class will never be increased again so we remove any occurrence
        # of the rule of any rule for that class from _rules_using_class and
        # _rules_pumping_class
        for rule_idx in self._rules_pumping_class[comb_class]:
            for child in self._rules[rule_idx].children:
                self._rules_using_class[child] = [
                    (ri, ci)
                    for ri, ci in self._rules_using_class[child]
                    if ri != rule_idx
                ]
        self._rules_pumping_class[comb_class].clear()
        # Correction of the shifts for rules using comb_class to pump
        for rule_idx, class_idx in self._rules_using_class[comb_class]:
            shifts = self._shifts[rule_idx]
            shifts[class_idx] = None
            if self._can_give_terms(shifts):
                self._processing_queue.append(rule_idx)
        self._rules_using_class[comb_class].clear()

    def rule_info(self, rule_idx: int) -> str:
        """
        Return a string with information about a particular rule.
        Mostly intended for debugging.
        """

        def v_to_str(v: Optional[int]) -> str:
            """Return a string for the integer and infinity if None"""
            if v is None:
                return "∞"
            return str(v)

        rule_key = self._rules[rule_idx]
        current_value = f"{v_to_str(self._function[rule_key.parent])} -> " + ", ".join(
            map(v_to_str, (self._function[c] for c in rule_key.children))
        )
        shifts = map(v_to_str, self._shifts[rule_idx])
        child_with_shift = ", ".join(
            f"({c}, {s})" for c, s in zip(rule_key.children, shifts)
        )
        return f"{rule_key.parent} -> {child_with_shift} || {current_value}"


class ForestRuleExtractor:
    MINIMIZE_ORDER = (
        RuleBucket.REVERSE,
        RuleBucket.NORMAL,
        RuleBucket.EQUIV,
        RuleBucket.VERIFICATION,
    )

    def __init__(
        self,
        root_label: int,
        ruledb: "RuleDBForest",
        classdb: ClassDB,
        pack: StrategyPack,
    ):
        self.pack = pack
        self.classdb = classdb
        self.root_label = root_label
        self.rule_by_bucket = self._sorted_stable_rules(ruledb.table_method)
        assert set(ForestRuleExtractor.MINIMIZE_ORDER) == set(self.rule_by_bucket)
        self.needed_rules: List[ForestRuleKey] = []
        self._minimize()

    def check(self) -> bool:
        """
        Make a sanity check of the status of the extractor.
        """
        lhs = set(rk.parent for rk in self.needed_rules)
        assert len(lhs) == len(self.needed_rules)
        assert self._is_productive(self.needed_rules)

    def rules(self, cache: Tuple[AbstractRule, ...]) -> Iterator[AbstractRule]:
        """
        Return all the rules of the specification.

        It will first try to use a rule in the cache and otherwise will try to
        recompute it from the pack.

        The empty rule are ignored as they be produced as needed by the specification.
        """
        cache_dict = {rule.forest_key(self.classdb.get_label): rule for rule in cache}
        for rk in self.needed_rules:
            if rk in cache_dict:
                rule = cache_dict[rk]
            else:
                rule = self._find_rule(rk)
            if isinstance(rule.strategy, EmptyStrategy):
                continue
            if rule.is_equivalence() and not isinstance(
                rule, (EquivalencePathRule, EquivalenceRule)
            ):
                assert isinstance(rule, Rule)
                yield rule.to_equivalence_rule()
            else:
                yield rule

    def _minimize(self):
        """
        Perform the complete minimization of the forest.
        """
        for key in ForestRuleExtractor.MINIMIZE_ORDER:
            self._minimize_key(key)

    def _minimize_key(self, key: RuleBucket) -> None:
        """
        Minimize the number of rules used for the type of rule given by key.

        The list of rule in `self.rule_by_bucket[key]` is cleared and a
        minimal set from theses is added to `self.needed_rules`.
        """
        logger.info("Minimizing %s", key.name)
        maybe_useful: List[ForestRuleKey] = []
        not_minimizing: List[List[ForestRuleKey]] = [
            self.needed_rules,
            maybe_useful,
        ]
        not_minimizing.extend(
            rules for k, rules in self.rule_by_bucket.items() if k != key
        )
        minimizing = self.rule_by_bucket[key]
        while minimizing:
            tb = TableMethod()
            # Add the rule we are not trying to minimize
            for rk in itertools.chain.from_iterable(not_minimizing):
                tb.add_rule_key(rk)
            if tb.is_pumping(self.root_label):
                minimizing.clear()
                break
            # Add rule until it gets productive
            for i, rk in enumerate(minimizing):
                tb.add_rule_key(rk)
                if tb.is_pumping(self.root_label):
                    break
            else:
                raise RuntimeError("Not pumping after adding all rules")
            maybe_useful.append(rk)
            assert minimizing, "variable i won't be set"
            # pylint: disable=undefined-loop-variable
            for _ in range(i, len(minimizing)):
                minimizing.pop()
        counter = 0
        while maybe_useful:
            rk = maybe_useful.pop()
            if not self._is_productive(itertools.chain.from_iterable(not_minimizing)):
                self.needed_rules.append(rk)
                counter += 1
        logger.info("Using %s rule for %s", counter, key.name)

    def _is_productive(self, rule_keys: Iterable[ForestRuleKey]) -> bool:
        """
        Check if the given set of rules is productive.
        """
        ruledb = TableMethod()
        for rk in rule_keys:
            ruledb.add_rule_key(rk)
        return ruledb.is_pumping(self.root_label)

    def _sorted_stable_rules(self, ruledb: TableMethod) -> SortedRWS:
        """
        Extract all the rule from the stable subuniverse and return all of them in a
        dict sorted by type.
        """
        res: SortedRWS = {bucket: [] for bucket in self.MINIMIZE_ORDER}
        for forest_key in ruledb.pumping_subuniverse():
            try:
                res[forest_key.bucket].append(forest_key)
            except KeyError as e:
                msg = (
                    f"{forest_key.bucket} type is not currently supported "
                    "by the extractor"
                )
                raise RuntimeError(msg) from e
        return res

    def _find_rule(self, rule_key: ForestRuleKey) -> AbstractRule:
        """
        Find a rule that have the given rule key.
        """
        all_classes = (rule_key.parent,) + rule_key.children
        all_normal_rules = itertools.chain.from_iterable(
            self._rules_for_class(c) for c in all_classes
        )
        for normal_rule in all_normal_rules:
            potential_rules = [normal_rule]
            if normal_rule.is_reversible():
                assert isinstance(normal_rule, Rule)
                potential_rules.extend(
                    normal_rule.to_reverse_rule(i)
                    for i in range(len(normal_rule.children))
                )
            for rule in potential_rules:
                if rule.forest_key(self.classdb.get_label) == rule_key:
                    return rule
        err = f"Can't find a rule for {rule_key}\n"
        err += f"Parent:\n{self.classdb.get_class(rule_key.parent)}\n"
        for i, l in enumerate(rule_key.children):
            err += f"Child {i}:\n{self.classdb.get_class(l)}\n"
        raise RuntimeError(err)

    def _rules_for_class(self, label: int) -> Iterator[AbstractRule]:
        """
        Return all the rule created for that class with the pack.
        """
        comb_class = self.classdb.get_class(label)
        strats: Iterator[CSSstrategy] = itertools.chain([EmptyStrategy()], self.pack)
        for strat in strats:
            if isinstance(strat, StrategyFactory):
                strats_or_rules: Iterable[
                    Union[AbstractRule, AbstractStrategy]
                ] = strat(comb_class)
            else:
                strats_or_rules = [strat]
            for x in strats_or_rules:
                if isinstance(x, AbstractStrategy):
                    try:
                        yield x(comb_class)
                    except StrategyDoesNotApply:
                        continue
                else:
                    yield x


class RuleDBForest(RuleDBAbstract):
    """
    The rule database that provides live information on which class are pumping with the
    current rule in the database.

    Set `reverse` to prevent the reverse of the added rules to be added to the database.
    """

    def __init__(
        self,
        *,
        reverse: bool = True,
        rule_cache: Iterable[AbstractRule] = tuple(),
    ) -> None:
        super().__init__()
        self.reverse = reverse
        self._num_rules = 0
        self._time_table_method = 0.0
        self._time_key = 0.0
        self.table_method = TableMethod()
        self._already_empty: Set[int] = set()
        self._rule_cache = tuple(rule_cache)

    # Implementation of RuleDBAbstract

    def status(self, elaborate: bool) -> str:
        s = "RuleDB status:\n"
        s += f"\tAdded from {self._num_rules} normal rules\n"
        key_time = timedelta(seconds=int(self._time_key))
        tm_time = timedelta(seconds=int(self._time_table_method))
        s += f"\tTime spent computing forest keys: {key_time}\n"
        s += f"\tTime spent running the table method: {tm_time}\n"
        s += self.table_method.status()
        return s

    def is_verified(self, label: int) -> bool:
        """
        Determine if the comb_class is pumping in the current universe.
        """
        return self.table_method.is_pumping(label)

    def has_specification(self) -> bool:
        return self.is_verified(self.root_label)

    def add(self, start: int, ends: Tuple[int, ...], rule: AbstractRule) -> None:
        self._add_empty_rule(ends, rule)
        self._num_rules += 1
        start_time = time.time()
        new_rule_keys = [rule.forest_key(self.classdb.get_label)]
        if self.reverse and rule.is_reversible():
            assert isinstance(rule, Rule)
            new_rule_keys.extend(
                rule.to_reverse_rule(i).forest_key(self.classdb.get_label)
                for i in range(len(rule.children))
            )
        self._time_key += time.time() - start_time
        start_time = time.time()
        for new_key in new_rule_keys:
            self.table_method.add_rule_key(new_key)
        self._time_table_method += time.time() - start_time

    @ensure_specification
    def get_specification_rules(self, **kwargs) -> Iterator[AbstractRule]:
        extractor = ForestRuleExtractor(
            self.root_label, self, self.classdb, self.strategy_pack
        )
        extractor.check()
        return extractor.rules(self._rule_cache)

    # Other methods

    def _add_empty_rule(self, ends: Iterable[int], rule: AbstractRule) -> None:
        """
        Add empty rule for the children of the rule if needed.
        """
        if not rule.possibly_empty:
            return
        for label, comb_class in zip(ends, rule.children):
            if label not in self._already_empty and self.classdb.is_empty(
                comb_class, label
            ):
                rule = empty_strategy(comb_class)
                self._already_empty.add(label)
                self.searcher.add_rule(label, (), rule)
